;;; Copyright (c) 2006, 2007 Abdulaziz Ghuloum and Kent Dybvig
;;; 
;;; Permission is hereby granted, free of charge, to any person obtaining a
;;; copy of this software and associated documentation files (the "Software"),
;;; to deal in the Software without restriction, including without limitation
;;; the rights to use, copy, modify, merge, publish, distribute, sublicense,
;;; and/or sell copies of the Software, and to permit persons to whom the
;;; Software is furnished to do so, subject to the following conditions:
;;; 
;;; The above copyright notice and this permission notice shall be included in
;;; all copies or substantial portions of the Software.
;;; 
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
;;; THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
;;; DEALINGS IN THE SOFTWARE. 

(import 
  (rnrs base)
  (rnrs control)
  (rnrs io simple)
  (rnrs io ports)
  (rnrs lists)
  (rnrs files)
  (rnrs hashtables)
  (psyntax internal)
  (psyntax compat)
  (psyntax library-manager)
  (psyntax expander)
  (system)
  (mosh)
  )


;; here we put libraries used in psyntax.
(define scheme-library-files
  `("psyntax/compat.ss"
    "psyntax/internal.ss"
    "psyntax/config.ss"
    "psyntax/library-manager.ss"
    "psyntax/builders.ss"
    "psyntax/expander.ss"
    ,(string-append (get-environment-variable "HOME") "/mosh/lib/mosh/condition.ss")
    ,(string-append (get-environment-variable "HOME") "/mosh/lib/mosh/io/conditions.ss")
    "psyntax/main.ss"))


(define psyntax-system-macros
  '((define              (define))
    (define-syntax       (define-syntax))
    (module              (module))
    (begin               (begin))
    (import              (import))
    (set!                (set!))
    (let-syntax          (let-syntax))
    (letrec-syntax       (letrec-syntax))
    (foreign-call        (core-macro . foreign-call))
    (quote               (core-macro . quote))
    (syntax-case         (core-macro . syntax-case))
    (syntax              (core-macro . syntax))
    (lambda              (core-macro . lambda))
    (case-lambda         (core-macro . case-lambda))
    (type-descriptor     (core-macro . type-descriptor))
    (letrec              (core-macro . letrec))
    (letrec*             (core-macro . letrec*))
    (if                  (core-macro . if))
    (when                (macro . when))         
    (unless              (macro . unless))
    (parameterize        (macro . parameterize))
    (case                (macro . case))
;;     (when                (core-macro . when))         
;;     (unless              (core-macro . unless))
;;     (parameterize        (core-macro . parameterize))
;;     (case                (core-macro . case))
;    (receive             (core-macro . receive))
    (record-type-descriptor (core-macro . record-type-descriptor))
    (record-constructor-descriptor (core-macro . record-constructor-descriptor))
    (define-struct       (macro . define-struct))
    (include             (macro . include))
    (syntax-rules        (macro . syntax-rules))
    (quasiquote          (macro . quasiquote))
    (quasisyntax         (macro . quasisyntax))
    (with-syntax         (macro . with-syntax))
    (identifier-syntax   (macro . identifier-syntax))
    (let-values          (macro . let-values))
    (let*-values          (macro . let*-values))
    (let                 (macro . let))
    (let*                (macro . let*))
    (cond                (macro . cond))
    (do                  (macro . do))
    (and                 (macro . and))
    (or                  (macro . or))
    (time                (macro . time))
    (delay               (macro . delay))
    (endianness          (macro . endianness))
    (assert              (macro . assert))
    (...                 (macro . ...))
    (=>                  (macro . =>))
    (else                (macro . else))
    (_                   (macro . _))
    (unquote             (macro . unquote))
    (unquote-splicing    (macro . unquote-splicing))
    (unsyntax            (macro . unsyntax))
    (unsyntax-splicing   (macro . unsyntax-splicing))
    (trace-lambda        (macro . trace-lambda))
    (trace-define        (macro . trace-define))
    ;;; new
    (guard                 (macro . guard))
    (eol-style             (macro . eol-style))
    (buffer-mode           (macro . buffer-mode))
    (file-options          (macro . file-options))
    (error-handling-mode   (macro . error-handling-mode))
    (fields                (macro . fields)) 
    (mutable               (macro . mutable))
    (immutable             (macro . immutable)) 
    (parent                (macro . parent))
    (protocol              (macro . protocol))
    (sealed                (macro . sealed))
    (opaque                (macro . opaque ))
    (nongenerative         (macro . nongenerative))
    (parent-rtd            (macro . parent-rtd))
    (define-record-type    (macro . define-record-type))
    (define-enumeration    (macro . define-enumeration))
    (define-condition-type (macro . define-condition-type))
    ;;; for (record-type-descriptor &condition-type) and
    ;;; (record-constructor-descriptor &condition-type) to
    ;;; expand properly, the implementation must export 
    ;;; the identifiers &condition-type-rtd, which must 
    ;;; be bound to the run-time value of the rtd, and 
    ;;; &condition-type-rcd which must be bound to the 
    ;;; corresponding record-constructor-descriptor.
    (&condition                ($core-rtd . (&condition-rtd &condition-rcd)))
    (&message                  ($core-rtd . (&message-rtd &message-rcd)))
    (&warning                  ($core-rtd . (&warning-rtd &warning-rcd )))
    (&serious                  ($core-rtd . (&serious-rtd &serious-rcd)))
    (&error                    ($core-rtd . (&error-rtd &error-rcd)))
    (&violation                ($core-rtd . (&violation-rtd &violation-rcd)))
    (&assertion                ($core-rtd . (&assertion-rtd &assertion-rcd)))
    (&irritants                ($core-rtd . (&irritants-rtd &irritants-rcd)))
    (&who                      ($core-rtd . (&who-rtd &who-rcd)))
    (&non-continuable          ($core-rtd . (&non-continuable-rtd &non-continuable-rcd)))
    (&implementation-restriction ($core-rtd . (&implementation-restriction-rtd &implementation-restriction-rcd)))
    (&lexical                  ($core-rtd . (&lexical-rtd &lexical-rcd)))
    (&syntax                   ($core-rtd . (&syntax-rtd &syntax-rcd)))
    (&undefined                ($core-rtd . (&undefined-rtd &undefined-rcd)))
    (&i/o                      ($core-rtd . (&i/o-rtd &i/o-rcd)))
    (&i/o-read                 ($core-rtd . (&i/o-read-rtd &i/o-read-rcd)))
    (&i/o-write                ($core-rtd . (&i/o-write-rtd &i/o-write-rcd)))
    (&i/o-invalid-position     ($core-rtd . (&i/o-invalid-position-rtd &i/o-invalid-position-rcd)))
    (&i/o-filename             ($core-rtd . (&i/o-filename-rtd &i/o-filename-rcd)))
    (&i/o-file-protection      ($core-rtd . (&i/o-file-protection-rtd &i/o-file-protection-rcd)))
    (&i/o-file-is-read-only    ($core-rtd . (&i/o-file-is-read-only-rtd &i/o-fie-is-read-only-rcd)))
    (&i/o-file-already-exists  ($core-rtd . (&i/o-file-already-exists-rtd &i/o-file-already-exists-rcd)))
    (&i/o-file-does-not-exist  ($core-rtd . (&i/o-file-does-not-exist-rtd &i/o-file-does-not-exist-rcd)))
    (&i/o-port                 ($core-rtd . (&i/o-port-rtd &i/o-port-rcd)))
    (&i/o-decoding             ($core-rtd . (&i/o-decoding-rtd &i/o-decoding-rcd)))
    (&i/o-encoding             ($core-rtd . (&i/o-encoding-rtd &i/o-encoding-rcd)))
    (&no-infinities            ($core-rtd . (&no-infinities-rtd &no-infinities-rcd)))
    (&no-nans                  ($core-rtd . (&no-nans-rtd &no-nans-rcd)))
    ))

(define (macro-identifier x)
  (assq x psyntax-system-macros))

(define library-legend
  ;; abbr.       name                             visible? required?
  '((interaction (mosh interaction)                  #t    #f)
    (ne          (psyntax null-environment-5)          #t    #f)
    (se          (psyntax scheme-report-environment-5) #t    #f)
    (cm          (psyntax modules)                     #t    #f)
    (parameters  (chez parameters)                     #t    #f)
    (r           (rnrs)                                #t    #t)
    (r5          (rnrs r5rs)                           #t    #t)
    (ct          (rnrs control)                        #t    #t)
    (ev          (rnrs eval)                           #t    #t)
    (mp          (rnrs mutable-pairs)                  #t    #t)
    (ms          (rnrs mutable-strings)                #t    #t)
    (pr          (rnrs programs)                       #t    #t)
    (sc          (rnrs syntax-case)                    #t    #t)
    (fi          (rnrs files)                          #t    #t)
    (sr          (rnrs sorting)                        #t    #t)
    (ba          (rnrs base)                           #t    #t)
    (ls          (rnrs lists)                          #t    #t)
    (is          (rnrs io simple)                      #t    #t)
    (bv          (rnrs bytevectors)                    #t    #t)
    (uc          (rnrs unicode)                        #t    #t)
    (ex          (rnrs exceptions)                     #t    #t)
    (bw          (rnrs arithmetic bitwise)             #t    #t)
    (fx          (rnrs arithmetic fixnums)             #t    #t)
    (fl          (rnrs arithmetic flonums)             #t    #t)
    (ht          (rnrs hashtables)                     #t    #t)
    (ip          (rnrs io ports)                       #t    #t)
    (en          (rnrs enums)                          #t    #t)
    (co          (rnrs conditions)                     #t    #t)
    (ri          (rnrs records inspection)             #t    #t)
    (rp          (rnrs records procedural)             #t    #t)
    (rs          (rnrs records syntactic)              #t    #t)
    ($all        (psyntax system $all)                 #f    #t)
    ($boot       (psyntax system $bootstrap)           #f    #t)
    (mosh        (mosh)                                #f    #t)  ;; for mosh
    (mosh-backend        (mosh backend)                #f    #t)  ;; for mosh
    (mosh-file   (mosh file)                           #f    #t)  ;; for mosh
    (mosh-bytevector (mosh bytevector)                 #f    #t)  ;; for mosh
    (mosh-number (mosh number)                         #f    #t)  ;; for mosh
    (mosh-string (mosh string)                         #f    #t)  ;; for mosh
    (mosh-list   (mosh list)                           #f    #t)  ;; for mosh
    (mosh-regexp (mosh regexp)                         #f    #t)  ;; for mosh
    (mosh-hashtable (mosh hashtable)                   #f    #t)  ;; for mosh
    (sys         (system)                              #f    #t)  ;; for mosh
    (srfi-1     (srfi :1)                              #f    #t)  ;; for mosh
    ))

;;; required? flag means that said library is required for 
;;; building the system.  The only non-r6rs required libraries
;;; should be (psyntax system $bootstrap) and (psyntax system $all).
;;; (psyntax system $bootstrap) should export, at a minimum, the 
;;; following procedures: gensym, symbol-value, set-symbol-value!,
;;; eval-core, and pretty-print.
;;; (psyntax system $all) is fabricated by the system to include 
;;; every identifier in the system.


(define identifier->library-map
  '(
    ;;;
    (import                                     interaction)
    (export                                     interaction)
    (library                                    interaction)
    (include                                    interaction mosh)
    (lambda                                     interaction r ba se ne)
    (and                                        interaction r ba se ne)
    (begin                                      interaction r ba se ne)
    (case                                       interaction r ba se ne)
    (cond                                       interaction r ba se ne)
    (define                                     interaction r ba se ne)
    (define-syntax                              interaction r ba se ne)
    (identifier-syntax                    interaction       r ba)
    (if                                    interaction      r ba se ne)
    (let                                    interaction     r ba se ne)
    (let*                                    interaction    r ba se ne)
    (let*-values                          interaction       r ba)
    (let-syntax                            interaction      r ba se ne)
    (let-values                           interaction       r ba)
    (letrec                                interaction      r ba se ne)
    (letrec*                              interaction       r ba)
    (letrec-syntax                         interaction      r ba se ne)
    (or                                     interaction     r ba se ne)
    (quasiquote                              interaction    r ba se ne)
    (quote                                    interaction   r ba se ne)
    (set!                                      interaction  r ba se ne)
    (syntax-rules                               interaction r ba se ne)
    (unquote                                    interaction r ba se ne)
    (unquote-splicing                           interaction r ba se ne)
    (<                                       interaction    r ba se)
    (<=                                      interaction    r ba se)
    (=                                       interaction    r ba se)
    (>                                       interaction    r ba se)
    (>=                                      interaction    r ba se)
    (+                                       interaction    r ba se)
    (-                                       interaction    r ba se)
    (*                                       interaction    r ba se)
    (/                                       interaction    r ba se)
    (abs                                     interaction    r ba se)
    (acos                                    interaction    r ba se)
    (angle                                   interaction    r ba se)
    (append                                  interaction    r ba se)
    (apply                                   interaction    r ba se)
    (asin                                    interaction    r ba se)
    (assert                               interaction       r ba)
    (assertion-violation                  interaction       r ba)
    (atan                                  interaction      r ba se)
    (boolean=?                            interaction       r ba)
    (boolean?                              interaction      r ba se)
    (car                                    interaction     r ba se)
    (cdr                                     interaction    r ba se)
    (caar                                    interaction    r ba se)
    (cadr                                    interaction    r ba se)
    (cdar                                    interaction    r ba se)
    (cddr                                    interaction    r ba se)
    (caaar                                   interaction    r ba se)
    (caadr                                   interaction    r ba se)
    (cadar                                   interaction    r ba se)
    (caddr                                   interaction    r ba se)
    (cdaar                                   interaction    r ba se)
    (cdadr                                   interaction    r ba se)
    (cddar                                   interaction    r ba se)
    (cdddr                                   interaction    r ba se)
    (caaaar                                  interaction    r ba se)
    (caaadr                                  interaction    r ba se)
    (caadar                                  interaction    r ba se)
    (caaddr                                  interaction    r ba se)
    (cadaar                                  interaction    r ba se)
    (cadadr                                  interaction    r ba se)
    (caddar                                  interaction    r ba se)
    (cadddr                                  interaction    r ba se)
    (cdaaar                                  interaction    r ba se)
    (cdaadr                                  interaction    r ba se)
    (cdadar                                  interaction    r ba se)
    (cdaddr                                  interaction    r ba se)
    (cddaar                                  interaction    r ba se)
    (cddadr                                  interaction    r ba se)
    (cdddar                                  interaction    r ba se)
    (cddddr                                  interaction    r ba se)
    (call-with-current-continuation          interaction    r ba se)
    (call/cc                              interaction       r ba)
    (call-with-values                      interaction      r ba se)
    (ceiling                                interaction     r ba se)
    (char->integer                           interaction    r ba se)
    (char<=?                                 interaction    r ba se)
    (char<?                                  interaction    r ba se)
    (char=?                                  interaction    r ba se)
    (char>=?                                 interaction    r ba se)
    (char>?                                  interaction    r ba se)
    (char?                                   interaction    r ba se)
    (complex?                                interaction    r ba se)
    (cons                                    interaction    r ba se)
    (cos                                     interaction    r ba se)
    (denominator                             interaction    r ba se)
    (div                                  interaction       r ba)
    (mod                                  interaction       r ba)
    (div-and-mod                          interaction       r ba)
    (div0                                 interaction       r ba)
    (mod0                                 interaction       r ba)
    (div0-and-mod0                        interaction       r ba)
    (dynamic-wind                          interaction      r ba se)
    (eq?                                    interaction     r ba se)
    (equal?                                  interaction    r ba se)
    (eqv?                                    interaction    r ba se)
    (error                                interaction       r ba)
    (even?                                 interaction      r ba se)
    (exact                                interaction       r ba)
    (exact-integer-sqrt                   interaction       r ba)
    (exact?                                interaction      r ba se)
    (exp                                    interaction     r ba se)
    (expt                                    interaction    r ba se)
    (finite?                              interaction       r ba)
    (floor                                 interaction      r ba se)
    (for-each                               interaction     r ba se)
    (gcd                                     interaction    r ba se)
    (imag-part                               interaction    r ba se)
    (inexact                              interaction       r ba)
    (inexact?                              interaction      r ba se)
    (infinite?                            interaction       r ba)
    (integer->char                         interaction      r ba se)
    (integer-valued?                      interaction       r ba)
    (integer?                              interaction      r ba se)
    (lcm                                    interaction     r ba se)
    (length                                  interaction    r ba se)
    (list                                    interaction    r ba se)
    (list->string                            interaction    r ba se)
    (list->vector                            interaction    r ba se)
    (list-ref                                interaction    r ba se)
    (list-tail                               interaction    r ba se)
    (list?                                   interaction    r ba se)
    (log                                     interaction    r ba se)
    (magnitude                               interaction    r ba se)
    (make-polar                              interaction    r ba se)
    (make-rectangular                        interaction    r ba se)
    (make-string                             interaction    r ba se)
    (make-vector                             interaction    r ba se)
    (map                                     interaction    r ba se)
    (max                                     interaction    r ba se)
    (min                                     interaction    r ba se)
    (nan?                                 interaction       r ba)
    (negative?                             interaction      r ba se)
    (not                                    interaction     r ba se)
    (null?                                interaction       r ba)
    (number->string                        interaction      r ba se)
    (number?                                interaction     r ba se)
    (numerator                               interaction    r ba se)
    (odd?                                    interaction    r ba se)
    (pair?                                   interaction    r ba se)
    (positive?                               interaction    r ba se)
    (procedure?                              interaction    r ba se)
    (rational-valued?                     interaction       r ba)
    (rational?                             interaction      r ba se)
    (rationalize                            interaction     r ba se)
    (real-part                               interaction    r ba se)
    (real-valued?                         interaction       r ba)
    (real?                                 interaction      r ba se)
    (reverse                                interaction     r ba se)
    (round                                   interaction    r ba se)
    (sin                                     interaction    r ba se)
    (sqrt                                    interaction    r ba se)
    (string                                  interaction    r ba se)
    (string->list                            interaction    r ba se)
    (string->number                          interaction    r ba se)
    (string->symbol                          interaction    r ba se)
    (string-append                           interaction    r ba se)
    (string-copy                             interaction    r ba se)
    (string-for-each                      interaction       r ba)
    (string-length                         interaction      r ba se)
    (string-ref                             interaction     r ba se)
    (string<=?                               interaction    r ba se)
    (string<?                                interaction    r ba se)
    (string=?                                interaction    r ba se)
    (string>=?                               interaction    r ba se)
    (string>?                                interaction    r ba se)
    (string?                                 interaction    r ba se)
    (substring                               interaction    r ba se)
    (symbol->string                          interaction    r ba se)
    (symbol=?                             interaction       r ba)
    (symbol?                               interaction      r ba se)
    (tan                                    interaction     r ba se)
    (truncate                                interaction    r ba se)
    (values                                  interaction    r ba se)
    (vector                                  interaction    r ba se)
    (vector->list                            interaction    r ba se)
    (vector-fill!                            interaction    r ba se)
    (vector-for-each                      interaction       r ba)
    (vector-length                         interaction      r ba se)
    (vector-map                           interaction       r ba)
    (vector-ref                            interaction      r ba se)
    (vector-set!                            interaction     r ba se)
    (vector?                                 interaction    r ba se)
    (zero?                                   interaction    r ba se)
    (...                                     interaction    r ba sc)
    (=>                                      interaction    r ba ex)
    (_                                       interaction    r ba sc)
    (else                                    interaction    r ba ex)
    ;;;
    (bitwise-arithmetic-shift                   interaction r bw)
    (bitwise-arithmetic-shift-left              interaction r bw)
    (bitwise-arithmetic-shift-right             interaction r bw)
    (bitwise-not                                interaction r bw)
    (bitwise-and                                interaction r bw)
    (bitwise-ior                                interaction r bw)
    (bitwise-xor                                interaction r bw)
    (bitwise-bit-count                          interaction r bw)
    (bitwise-bit-field                          interaction r bw)
    (bitwise-bit-set?                           interaction r bw)
    (bitwise-copy-bit                           interaction r bw)
    (bitwise-copy-bit-field                     interaction r bw)
    (bitwise-first-bit-set                      interaction r bw)
    (bitwise-if                                 interaction r bw)
    (bitwise-length                             interaction r bw)
    (bitwise-reverse-bit-field                  interaction r bw)
    (bitwise-rotate-bit-field                   interaction r bw)
    ;;;
    (fixnum?                                    interaction r fx)
    (fixnum-width                               interaction r fx) 
    (least-fixnum                               interaction r fx)
    (greatest-fixnum                            interaction r fx) 
    (fx*                                        interaction r fx)
    (fx*/carry                                  interaction r fx)
    (fx+                                        interaction r fx)
    (fx+/carry                                  interaction r fx)
    (fx-                                        interaction r fx)
    (fx-/carry                                  interaction r fx)
    (fx<=?                                      interaction r fx)
    (fx<?                                       interaction r fx)
    (fx=?                                       interaction r fx)
    (fx>=?                                      interaction r fx)
    (fx>?                                       interaction r fx)
    (fxand                                      interaction r fx)
    (fxarithmetic-shift                         interaction r fx)
    (fxarithmetic-shift-left                    interaction r fx)
    (fxarithmetic-shift-right                   interaction r fx)
    (fxbit-count                                interaction r fx)
    (fxbit-field                                interaction r fx)
    (fxbit-set?                                 interaction r fx)
    (fxcopy-bit                                 interaction r fx)
    (fxcopy-bit-field                           interaction r fx)
    (fxdiv                                      interaction r fx)
    (fxdiv-and-mod                              interaction r fx)
    (fxdiv0                                     interaction r fx)
    (fxdiv0-and-mod0                            interaction r fx)
    (fxeven?                                    interaction r fx)
    (fxfirst-bit-set                            interaction r fx)
    (fxif                                       interaction r fx)
    (fxior                                      interaction r fx)
    (fxlength                                   interaction r fx)
    (fxmax                                      interaction r fx)
    (fxmin                                      interaction r fx)
    (fxmod                                      interaction r fx)
    (fxmod0                                     interaction r fx)
    (fxnegative?                                interaction r fx)
    (fxnot                                      interaction r fx)
    (fxodd?                                     interaction r fx)
    (fxpositive?                                interaction r fx)
    (fxreverse-bit-field                        interaction r fx)
    (fxrotate-bit-field                         interaction r fx)
    (fxxor                                      interaction r fx)
    (fxzero?                                    interaction r fx)
;;;
    (fixnum->flonum                             interaction r fl)
    (fl*                                        interaction r fl)
    (fl+                                        interaction r fl)
    (fl-                                        interaction r fl)
    (fl/                                        interaction r fl)
    (fl<=?                                      interaction r fl)
    (fl<?                                       interaction r fl)
    (fl=?                                       interaction r fl)
    (fl>=?                                      interaction r fl)
    (fl>?                                       interaction r fl)
    (flabs                                      interaction r fl)
    (flacos                                     interaction r fl)
    (flasin                                     interaction r fl)
    (flatan                                     interaction r fl)
    (flceiling                                  interaction r fl)
    (flcos                                      interaction r fl)
    (fldenominator                              interaction r fl)
    (fldiv                                      interaction r fl)
    (fldiv-and-mod                              interaction r fl)
    (fldiv0                                     interaction r fl)
    (fldiv0-and-mod0                            interaction r fl)
    (fleven?                                    interaction r fl)
    (flexp                                      interaction r fl)
    (flexpt                                     interaction r fl)
    (flfinite?                                  interaction r fl)
    (flfloor                                    interaction r fl)
    (flinfinite?                                interaction r fl)
    (flinteger?                                 interaction r fl)
    (fllog                                      interaction r fl)
    (flmax                                      interaction r fl)
    (flmin                                      interaction r fl)
    (flmod                                      interaction r fl)
    (flmod0                                     interaction r fl)
    (flnan?                                     interaction r fl)
    (flnegative?                                interaction r fl)
    (flnumerator                                interaction r fl)
    (flodd?                                     interaction r fl)
    (flonum?                                    interaction r fl)
    (flpositive?                                interaction r fl)
    (flround                                    interaction r fl)
    (flsin                                      interaction r fl)
    (flsqrt                                     interaction r fl)
    (fltan                                      interaction r fl)
    (fltruncate                                 interaction r fl)
    (flzero?                                    interaction r fl)
    (real->flonum                               interaction r fl)
    (make-no-infinities-violation               interaction r fl)
    (make-no-nans-violation                     interaction r fl)
    (&no-infinities                             interaction r fl)
    (no-infinities-violation?                   interaction r fl)
    (&no-nans                                   interaction r fl)
    (no-nans-violation?                         interaction r fl)
    ;;;
    (bytevector->sint-list                      interaction r bv)
    (bytevector->u8-list                        interaction r bv)
    (bytevector->uint-list                      interaction r bv)
    (bytevector-copy                            interaction r bv)
    (bytevector-copy!                           interaction r bv)
    (bytevector-fill!                           interaction r bv)
    (bytevector-ieee-double-native-ref          interaction r bv)
    (bytevector-ieee-double-native-set!         interaction r bv)
    (bytevector-ieee-double-ref                 interaction r bv)
    (bytevector-ieee-single-native-ref          interaction r bv)
    (bytevector-ieee-single-native-set!         interaction r bv)
    (bytevector-ieee-single-ref                 interaction r bv)
    (bytevector-length                          interaction r bv)
    (bytevector-s16-native-ref                  interaction r bv)
    (bytevector-s16-native-set!                 interaction r bv)
    (bytevector-s16-ref                         interaction r bv)
    (bytevector-s16-set!                        interaction r bv)
    (bytevector-s32-native-ref                  interaction r bv)
    (bytevector-s32-native-set!                 interaction r bv)
    (bytevector-s32-ref                         interaction r bv)
    (bytevector-s32-set!                        interaction r bv)
    (bytevector-s64-native-ref                  interaction r bv)
    (bytevector-s64-native-set!                 interaction r bv)
    (bytevector-s64-ref                         interaction r bv)
    (bytevector-s64-set!                        interaction r bv)
    (bytevector-s8-ref                          interaction r bv)
    (bytevector-s8-set!                         interaction r bv)
    (bytevector-sint-ref                        interaction r bv)
    (bytevector-sint-set!                       interaction r bv)
    (bytevector-u16-native-ref                  interaction r bv)
    (bytevector-u16-native-set!                 interaction r bv)
    (bytevector-u16-ref                         interaction r bv)
    (bytevector-u16-set!                        interaction r bv)
    (bytevector-u32-native-ref                  interaction r bv)
    (bytevector-u32-native-set!                 interaction r bv)
    (bytevector-u32-ref                         interaction r bv)
    (bytevector-u32-set!                        interaction r bv)
    (bytevector-u64-native-ref                  interaction r bv)
    (bytevector-u64-native-set!                 interaction r bv)
    (bytevector-u64-ref                         interaction r bv)
    (bytevector-u64-set!                        interaction r bv)
    (bytevector-u8-ref                          interaction r bv)
    (bytevector-u8-set!                         interaction r bv)
    (bytevector-uint-ref                        interaction r bv)
    (bytevector-uint-set!                       interaction r bv)
    (bytevector=?                               interaction r bv)
    (bytevector?                                interaction r bv)
    (endianness                                 interaction r bv)
    (native-endianness                          interaction r bv)
    (sint-list->bytevector                      interaction r bv)
    (string->utf16                              interaction r bv)
    (string->utf32                              interaction r bv)
    (string->utf8                               interaction r bv)
    (u8-list->bytevector                        interaction r bv)
    (uint-list->bytevector                      interaction r bv)
    (utf8->string                               interaction r bv)
    (utf16->string                              interaction r bv)
    (utf32->string                              interaction r bv)
;;;
    (condition?                                 interaction r co)
    (&assertion                                 interaction r co)
    (assertion-violation?                       interaction r co)
    (&condition                                 interaction r co)
    (condition                                  interaction r co)
    (condition-accessor                         interaction r co)
    (condition-irritants                        interaction r co)
    (condition-message                          interaction r co)
    (condition-predicate                        interaction r co)
    (condition-who                              interaction r co)
    (define-condition-type                      interaction r co)
    (&error                                     interaction r co)
    (error?                                     interaction r co)
    (&implementation-restriction                interaction r co)
    (implementation-restriction-violation?      interaction r co)
    (&irritants                                 interaction r co)
    (irritants-condition?                       interaction r co)
    (&lexical                                   interaction r co)
    (lexical-violation?                         interaction r co)
    (make-assertion-violation                   interaction r co)
    (make-error                                 interaction r co)
    (make-implementation-restriction-violation  interaction r co)
    (make-irritants-condition                   interaction r co)
    (make-lexical-violation                     interaction r co)
    (make-message-condition                     interaction r co)
    (make-non-continuable-violation             interaction r co)
    (make-serious-condition                     interaction r co)
    (make-syntax-violation                      interaction r co)
    (make-undefined-violation                   interaction r co)
    (make-violation                             interaction r co)
    (make-warning                               interaction r co)
    (make-who-condition                         interaction r co)
    (&message                                   interaction r co)
    (message-condition?                         interaction r co)
    (&non-continuable                           interaction r co)
    (non-continuable-violation?                 interaction r co)
    (&serious                                   interaction r co)
    (serious-condition?                         interaction r co)
    (simple-conditions                          interaction r co)
    (&syntax                                    interaction r co)
    (syntax-violation                           interaction r co sc)
    (syntax-violation-form                      interaction r co)
    (syntax-violation-subform                   interaction r co)
    (syntax-violation?                          interaction r co)
    (&undefined                                 interaction r co)
    (undefined-violation?                       interaction r co)
    (&violation                                 interaction r co)
    (violation?                                 interaction r co)
    (&warning                                   interaction r co)
    (warning?                                   interaction r co)
    (&who                                       interaction r co)
    (who-condition?                             interaction r co)
    ;;;
    (case-lambda                                interaction r ct)
    (do                                         interaction r ct se ne)
    (unless                                     interaction r ct)
    (when                                       interaction r ct)
    ;;;
    (define-enumeration                         interaction r en)
    (enum-set->list                             interaction r en)
    (enum-set-complement                        interaction r en)
    (enum-set-constructor                       interaction r en)
    (enum-set-difference                        interaction r en)
    (enum-set-indexer                           interaction r en)
    (enum-set-intersection                      interaction r en)
    (enum-set-member?                           interaction r en)
    (enum-set-projection                        interaction r en)
    (enum-set-subset?                           interaction r en)
    (enum-set-union                             interaction r en)
    (enum-set-universe                          interaction r en)
    (enum-set=?                                 interaction r en)
    (make-enumeration                           interaction r en)
    ;interaction ;;
    (environment                                interaction ev)
    (eval                                       interaction ev se)
    ;;;
    (raise                                      interaction r ex)
    (raise-continuable                          interaction r ex)
    (with-exception-handler                     interaction r ex)
    (guard                                      interaction r ex)
    ;;;
    (binary-port?                               interaction r ip)
    (buffer-mode                                interaction r ip)
    (buffer-mode?                               interaction r ip)
    (bytevector->string                         interaction r ip)
    (call-with-bytevector-output-port           interaction r ip)
    (call-with-port                             interaction r ip)
    (call-with-string-output-port               interaction r ip)
    ;;;
    (assoc                                      interaction r ls se)
    (assp                                       interaction r ls)
    (assq                                       interaction r ls se)
    (assv                                       interaction r ls se)
    (cons*                                      interaction r ls)
    (filter                                     interaction r ls)
    (find                                       interaction r ls)
    (fold-left                                  interaction r ls)
    (fold-right                                 interaction r ls)
    (for-all                                    interaction r ls)
    (exists                                     interaction r ls)
    (member                                     interaction r ls se)
    (memp                                       interaction r ls)
    (memq                                       interaction r ls se)
    (memv                                       interaction r ls se)
    (partition                                  interaction r ls)
    (remq                                       interaction r ls)
    (remp                                       interaction r ls)
    (remv                                       interaction r ls)
    (remove                                     interaction r ls)
    ;;;
    (set-car!                                   interaction mp se)
    (set-cdr!                                   interaction mp se)
    ;;;
    (string-set!                                interaction ms se)
    (string-fill!                               interaction ms se)
    ;;;
    (command-line                               interaction r pr)
    (exit                                       interaction r pr)
    ;;;
    (delay                                      r5 se ne)
    (exact->inexact                             r5 se)
    (force                                      r5 se)
    (inexact->exact                             r5 se)
    (modulo                                     r5 se)
    (remainder                                  r5 se)
    (null-environment                           r5 se)
    (quotient                                   r5 se)
    (scheme-report-environment                  r5 se)
    ;;;
    (close-port                                 interaction r ip)
    (eol-style                                  interaction r ip)
    (error-handling-mode                        interaction r ip)
    (file-options                               interaction r ip)
    (flush-output-port                          interaction r ip)
    (get-bytevector-all                         interaction r ip)
    (get-bytevector-n                           interaction r ip)
    (get-bytevector-n!                          interaction r ip)
    (get-bytevector-some                        interaction r ip)
    (get-char                                   interaction r ip)
    (get-datum                                  interaction r ip)
    (get-line                                   interaction r ip)
    (get-string-all                             interaction r ip)
    (get-string-n                               interaction r ip)
    (get-string-n!                              interaction r ip)
    (get-u8                                     interaction r ip)
    (&i/o                                       interaction r ip is fi)
    (&i/o-decoding                              interaction r ip)
    (i/o-decoding-error?                        interaction r ip)
    (&i/o-encoding                              interaction r ip)
    (i/o-encoding-error-char                    interaction r ip)
    (i/o-encoding-error?                        interaction r ip)
    (i/o-error-filename                         interaction r ip is fi)
    (i/o-error-port                             interaction r ip is fi)
    (i/o-error?                                 interaction r ip is fi)
    (&i/o-file-already-exists                   interaction r ip is fi)
    (i/o-file-already-exists-error?             interaction r ip is fi)
    (&i/o-file-does-not-exist                   interaction r ip is fi)
    (i/o-file-does-not-exist-error?             interaction r ip is fi)
    (&i/o-file-is-read-only                     interaction r ip is fi)
    (i/o-file-is-read-only-error?               interaction r ip is fi)
    (&i/o-file-protection                       interaction r ip is fi)
    (i/o-file-protection-error?                 interaction r ip is fi)
    (&i/o-filename                              interaction r ip is fi)
    (i/o-filename-error?                        interaction r ip is fi)
    (&i/o-invalid-position                      interaction r ip is fi)
    (i/o-invalid-position-error?                interaction r ip is fi)
    (&i/o-port                                  interaction r ip is fi)
    (i/o-port-error?                            interaction r ip is fi)
    (&i/o-read                                  interaction r ip is fi)
    (i/o-read-error?                            interaction r ip is fi)
    (&i/o-write                                 interaction r ip is fi)
    (i/o-write-error?                           interaction r ip is fi)
    (lookahead-char                             interaction r ip)
    (lookahead-u8                               r interaction  ip)
    (make-bytevector                            r interaction  bv)
    (make-custom-binary-input-port              r interaction  ip)
    (make-custom-binary-input/output-port       r interaction  ip)
    (make-custom-binary-output-port             r interaction  ip)
    (make-custom-textual-input-port             r interaction  ip)
    (make-custom-textual-input/output-port      r interaction  ip)
    (make-custom-textual-output-port            r interaction  ip)
    (make-i/o-decoding-error                    r interaction  ip)
    (make-i/o-encoding-error                    r interaction  ip)
    (make-i/o-error                             r interaction  ip is fi)
    (make-i/o-file-already-exists-error         r interaction  ip is fi)
    (make-i/o-file-does-not-exist-error         r interaction  ip is fi)
    (make-i/o-file-is-read-only-error           r interaction  ip is fi)
    (make-i/o-file-protection-error             r interaction  ip is fi)
    (make-i/o-filename-error                    r interaction  ip is fi)
    (make-i/o-invalid-position-error            r interaction  ip is fi)
    (make-i/o-port-error                        r interaction  ip is fi)
    (make-i/o-read-error                        r interaction  ip is fi)
    (make-i/o-write-error                       r interaction  ip is fi)
    (latin-1-codec                              r interaction  ip)
    (make-transcoder                            r interaction  ip)
    (native-eol-style                           r interaction  ip)
    (native-transcoder                          r interaction  ip)
    (open-bytevector-input-port                 r interaction  ip)
    (open-bytevector-output-port                r interaction  ip)
    (open-file-input-port                       r interaction  ip)
    (open-file-input/output-port                r interaction  ip)
    (open-file-output-port                      r interaction  ip)
    (open-string-input-port                     r interaction  ip)
    (open-string-output-port                    r interaction  ip)
    (output-port-buffer-mode                    r interaction  ip)
    (port-eof?                                  r interaction  ip)
    (port-has-port-position?                    r interaction  ip)
    (port-has-set-port-position!?               r interaction  ip)
    (port-position                              r interaction  ip)
    (port-transcoder                            r interaction  ip)
    (port?                                      r interaction  ip)
    (put-bytevector                             r interaction  ip)
    (put-char                                   r interaction  ip)
    (put-datum                                  r interaction  ip)
    (put-string                                 r interaction  ip)
    (put-u8                                     r interaction  ip)
    (set-port-position!                         r interaction  ip)
    (standard-error-port                        r interaction  ip)
    (standard-input-port                        r interaction  ip)
    (standard-output-port                       r interaction  ip)
    (string->bytevector                         r interaction  ip)
    (textual-port?                              r interaction  ip)
    (transcoded-port                            r interaction  ip)
    (transcoder-codec                           r interaction  ip)
    (transcoder-eol-style                       r interaction  ip)
    (transcoder-error-handling-mode             r interaction  ip)
    (utf-16-codec                               r interaction  ip)
    (utf-8-codec                                r interaction  ip)
    ;;;
    (input-port?                                r interaction  is ip se)
    (output-port?                               r interaction  is ip se)
    (current-input-port                         r interaction  ip is se)
    (current-output-port                        r interaction  ip is se)
    (current-error-port                         r interaction  ip is)
    (eof-object                                 r interaction  ip is se)
    (eof-object?                                r interaction  ip is)
    (close-input-port                           r interaction  is se)
    (close-output-port                          r interaction  is se)
    (display                                    r interaction  is se)
    (newline                                    r interaction  is se)
    (open-input-file                            r interaction  is se)
    (open-output-file                           r interaction  is se)
    (peek-char                                  r interaction  is se)
    (read                                       r interaction  is se)
    (read-char                                  r interaction  is se)
    (with-input-from-file                       r interaction  is se)
    (with-output-to-file                        r interaction  is se)
    (write                                      r interaction  is se)
    (write-char                                 r interaction  is se)
    (call-with-input-file                       r interaction  is se)
    (call-with-output-file                      r interaction  is se)
    ;;;
    (hashtable-clear!                           r interaction  ht)
    (hashtable-contains?                        r interaction  ht)
    (hashtable-copy                             r interaction  ht)
    (hashtable-delete!                          r interaction  ht)
    (hashtable-entries                          r interaction  ht)
    (hashtable-keys                             r interaction  ht)
    (hashtable-mutable?                         r interaction  ht)
    (hashtable-ref                              r interaction  ht)
    (hashtable-set!                             r interaction  ht)
    (hashtable-size                             r interaction  ht)
    (hashtable-update!                          r interaction  ht)
    (hashtable?                                 r interaction  ht)
    (make-eq-hashtable                          r interaction  ht)
    (make-eqv-hashtable                         r interaction  ht)
    (hashtable-hash-function                    r interaction  ht)
    (make-hashtable                             r interaction  ht)
    (hashtable-equivalence-function             r interaction  ht)
    (equal-hash                                 r interaction  ht)
    (string-hash                                r interaction  ht)
    (string-ci-hash                             r interaction  ht)
    (symbol-hash                                r interaction  ht)
    ;;;
    (list-sort                                  r interaction  sr)
    (vector-sort                                r interaction  sr)
    (vector-sort!                               r interaction  sr)
    ;;;
    (file-exists?                               r interaction  fi)
    (delete-file                                r interaction  fi)
    ;;;
    (define-record-type                         r interaction rs)
    (fields                                     r interaction rs)
    (immutable                                  r interaction rs)
    (mutable                                    r interaction rs)
    (opaque                                     r interaction rs)
    (parent                                     r interaction rs)
    (parent-rtd                                 r interaction rs)
    (protocol                                   r interaction rs)
    (record-constructor-descriptor              r interaction rs)
    (record-type-descriptor                     r interaction rs)
    (sealed                                     r interaction rs)
    (nongenerative                              r interaction rs)
    ;;;
    (record-field-mutable?                      r interaction ri)
    (record-rtd                                 r interaction ri)
    (record-type-field-names                    r interaction ri)
    (record-type-generative?                    r interaction ri)
    (record-type-name                           r interaction ri)
    (record-type-opaque?                        r interaction ri)
    (record-type-parent                         r interaction ri)
    (record-type-sealed?                        r interaction ri)
    (record-type-uid                            r interaction ri)
    (record?                                    r interaction ri)
    ;;;
    (make-record-constructor-descriptor         r interaction rp)
    (make-record-type-descriptor                r interaction rp)
    (record-accessor                            r interaction rp)
    (record-constructor                         r interaction rp)
    (record-mutator                             r interaction rp)
    (record-predicate                           r interaction rp)
    (record-type-descriptor?                    r interaction rp)
    ;;;
    (bound-identifier=?                         r interaction  sc)
    (datum->syntax                              r interaction  sc)
    (syntax                                     r interaction  sc)
    (syntax->datum                              r interaction  sc)
    (syntax-case                                r interaction  sc)
    (unsyntax                                   r interaction  sc)
    (unsyntax-splicing                          r interaction  sc)
    (quasisyntax                                r interaction  sc)
    (with-syntax                                r interaction  sc)
    (free-identifier=?                          r interaction  sc)
    (generate-temporaries                       r interaction  sc)
    (identifier?                                r interaction  sc)
    (make-variable-transformer                  r interaction  sc)
    ;;;
    (char-alphabetic?                           r interaction  uc se)
    (char-ci<=?                                 r interaction  uc se)
    (char-ci<?                                  r interaction  uc se)
    (char-ci=?                                  r interaction  uc se)
    (char-ci>=?                                 r interaction  uc se)
    (char-ci>?                                  r interaction  uc se)
    (char-downcase                              r interaction  uc se)
    (char-foldcase                              r interaction  uc)
    (char-titlecase                             r interaction  uc)
    (char-upcase                                r interaction  uc se)
    (char-general-category                      r interaction  uc)
    (char-lower-case?                           r interaction  uc se)
    (char-numeric?                              r interaction  uc se)
    (char-title-case?                           r interaction  uc)
    (char-upper-case?                           r interaction  uc se)
    (char-whitespace?                           r interaction  uc se)
    (string-ci<=?                               r interaction  uc se)
    (string-ci<?                                r interaction  uc se)
    (string-ci=?                                r interaction  uc se)
    (string-ci>=?                               r interaction  uc se)
    (string-ci>?                                r interaction  uc se)
    (string-downcase                            r interaction  uc)
    (string-foldcase                            r interaction  uc)
    (string-normalize-nfc                       r interaction  uc)
    (string-normalize-nfd                       r interaction  uc)
    (string-normalize-nfkc                      r interaction  uc)
    (string-normalize-nfkd                      r interaction  uc)
    (string-titlecase                           r interaction  uc)
    (string-upcase                              r interaction  uc)
    ;;; mosh
    ;;; string-split など操作の対象が string で戻り値も string に関連するものなら
    ;;; => (mosh string)
    ;;; readdir のようにファイル操作に関わるものは
    ;;; => (mosh file)
    ;;; string->regexp のように型を変換するものは source の string の型に合わせたライブラリに
    ;;; => (mosh string)
    ;;;  操作の中心的な対象が何であるか？で所属のライブラリを決める
    (receive mosh-backend) ;; I want psyntax pass through "receive" to backend(Mosh).
    (fasl-write mosh)
    (fasl-read mosh)
    (set-source-info! mosh)
    (make-instruction mosh)
    (make-compiler-instruction mosh)
    (source-info mosh)
    (sys-display mosh)
    (get-command-line mosh)
    (get-environment-variable sys)
    (get-environment-variables sys)
    (readdir sys)
    (hashtable-for-each mosh-hashtable)
    (regexp-replace-all mosh-regexp)
    (rxmatch mosh-regexp)
    (string->regexp mosh-string)
    (bytevector-for-each mosh-bytevector)
    (string-split mosh-string)
    (call-with-string-io mosh-string)
    (call-with-string-input-port mosh-string)
    (digit->integer mosh-number)
    (file->string mosh-file)
    (file->list mosh-file)
    (write-to-file mosh-file)
    (current-directory mosh-file)
    (stat-mtime mosh-file)
    (file-newer? mosh-file)
    (standard-library-path mosh)
    (format mosh-string)
    (print mosh-string)
    (alist-cons srfi-1)
    (assoc-ref mosh-list)
    (first srfi-1)
    (second srfi-1)
    (third srfi-1)
    (car+cdr srfi-1)
    (take srfi-1)
    (drop srfi-1)
    (take! srfi-1)
    (take-right srfi-1)
    (drop-right srfi-1)
    (drop-right! srfi-1)
    (fifth srfi-1)
    (sixth srfi-1)
    (seventh srfi-1)
    (eighth srfi-1)
    (ninth srfi-1)
    (tenth srfi-1)
    (xcons srfi-1)
    (make-list srfi-1)
    (list-tabulate srfi-1)
    (list-copy srfi-1)
    (list= srfi-1)
    (split-at srfi-1)
    (split-at! srfi-1)
    (not-pair? srfi-1)
    (last-pair srfi-1)
    (last srfi-1)
    (circular-list srfi-1)
    (proper-list? srfi-1)
    (circular-list? srfi-1)
    (dotted-list? srfi-1)
    (read-line mosh-file)
    (regexp? r)
    ;;;
    (char-ready?                                )
    (interaction-environment                    )
    (load                                       )
    ;;;
    (void                     $boot)
    (gensym                   $boot)
    (symbol-value             $boot mosh)
    (set-symbol-value!        $boot mosh)
    (eval-core                $boot)
    (pretty-print             $boot)
    (module                   cm)
    (syntax-dispatch ) ; only goes to $all
    (syntax-error    ) ; only goes to $all

    ;; mosh
    (&condition-rtd)
    (&condition-rcd)
    (&message-rtd)
    (&message-rcd)
    (&warning-rtd)
    (&warning-rcd)
    (&serious-rtd)
    (&serious-rcd)
    (&error-rtd)
    (&error-rcd)
    (&violation-rtd)
    (&violation-rcd)
    (&assertion-rtd)
    (&assertion-rcd)
    (&irritants-rtd)
    (&irritants-rcd)
    (&who-rtd)
    (&who-rcd)
    (&non-continuable-rtd)
    (&non-continuable-rcd)
    (&implementation-restriction-rtd)
    (&implementation-restriction-rcd)
    (&lexical-rtd)
    (&lexical-rcd)
    (&syntax-rtd)
    (&syntax-rcd)
    (&undefined-rtd)
    (&undefined-rcd)
    (&i/o-rtd)
    (&i/o-rcd)
    (&i/o-read-rtd)
    (&i/o-read-rcd)
    (&i/o-write-rtd)
    (&i/o-write-rcd)
    (&i/o-invalid-position-rtd)
    (&i/o-invalid-position-rcd)
    (&i/o-filename-rtd)
    (&i/o-filename-rcd)
    (&i/o-file-protection-rtd)
    (&i/o-file-protection-rcd)
    (&i/o-file-is-read-only-rtd)
    (&i/o-file-is-read-only-rcd)
    (&i/o-file-already-exists-rtd)
    (&i/o-file-already-exists-rcd)
    (&i/o-file-does-not-exist-rtd)
    (&i/o-file-does-not-exist-rcd)
    (&i/o-port-rtd)
    (&i/o-port-rcd)
    (&i/o-decoding-rtd)
    (&i/o-decoding-rcd)
    (&i/o-encoding-rtd)
    (&i/o-encoding-rcd)
    (&no-infinities-rtd)
    (&no-infinities-rcd)
    (&no-nans-rtd)
    (&no-nans-rcd)
    ))


(define (verify-map)
  (define (f x)
    (for-each
      (lambda (x)
        (unless (assq x library-legend)
          (error 'verify "~s is not in the libraries list" x)))
      (cdr x)))
  (for-each f identifier->library-map))

(define (make-collection)
  (let ((set '()))
    (case-lambda
      (() set)
      ((x) (set! set (cons x set))))))


(define (make-system-data subst env)
  (define who 'make-system-data)
  (let ((export-subst    (make-collection))
        (export-env      (make-collection))
        (export-primlocs (make-collection)))
    (for-each
      (lambda (x)
        (let ((name (car x)) (binding (cadr x)))
          (let ((label (gensym)))
            (export-subst (cons name label))
            (export-env   (cons label binding)))))
      psyntax-system-macros)
    (for-each
      (lambda (x)
        (cond
          ((macro-identifier x) (values))
          ((assq x (export-subst))
           (error who "ambiguous export of ~s" x))
          ((assq x subst) =>
           ;;; primitive defined (exported) within the compiled libraries
           (lambda (p)
             (let ((label (cdr p)))
               (cond
                 ((assq label env) =>
                  (lambda (p)
                    (let ((binding (cdr p)))
                      (case (car binding)
                        ((global)
                         (export-subst (cons x label))
                         (export-env   (cons label (cons 'core-prim x)))
                         (export-primlocs (cons x (cdr binding))))
                        (else
                         (error #f "invalid binding ~s for ~s" p x))))))
                 (else
                  (error #f "cannot find binding for ~s" x))))))
          (else
           ;;; core primitive with no backing definition, assumed to
           ;;; be defined in other strata of the system
           (let ((label (gensym)))
             (export-subst (cons x label))
             (export-env (cons label (cons 'core-prim x)))))))
      (map car identifier->library-map))
    (values (export-subst) (export-env) (export-primlocs))))

(define identifier->library-map-hashtable
  (let ((ht (make-eq-hashtable)))
    (for-each 
      (lambda (x)
        (hashtable-set! ht (car x) x))
      identifier->library-map)
    ht))

(define (get-export-subset key subst)
  (let f ((ls subst))
    (cond
      ((null? ls) '())
      (else
       (let ((x (car ls)))
         (let ((name (car x)))
           (cond
             ((hashtable-ref identifier->library-map-hashtable name #f);(assq name identifier->library-map)
              =>
              (lambda (q)
                (cond
                  ((memq key (cdr q))
                   (cons x (f (cdr ls))))
                  (else (f (cdr ls))))))
             (else
              ;;; not going to any library?
              (f (cdr ls))))))))))

(define (build-system-library export-subst export-env primlocs)
  (define (build-library legend-entry)
    (let ((key (car legend-entry))
          (name (cadr legend-entry))
          (visible? (caddr legend-entry)))
      (let ((id     (gensym))
            (name       name)
            (version     (if (eq? (car name) 'rnrs) '(6) '()))
            (import-libs '())
            (visit-libs  '())
            (invoke-libs '()))
        (let-values (((subst env)
                      (if (equal? name '(psyntax system $all))
                          (values export-subst export-env)
                          (values
                            (get-export-subset key export-subst)
                            '()))))
          `(install-library
             ',id ',name ',version ',import-libs ',visit-libs ',invoke-libs
             ',subst ',env values values '#f '#f ',visible? '#f)))))
  (let ((code `(library (psyntax primlocs)
                  (export) ;;; must be empty
                  (import
                    (only (psyntax library-manager)
                          install-library)
                    (only (psyntax internal)
                          current-primitive-locations)
                    (rnrs lists)
                    (rnrs base))
                  (current-primitive-locations
                    (lambda (x)
                      (cond
                        ((assq x ',primlocs) => cdr)
                        (else #f))))
                  ,@(map build-library library-legend))))
    (let-values (((name code empty-subst empty-env)
                  (boot-library-expand code)))
       code)))

(define (make-init-code)
  (values '() '() '()))

(define (expand-all files)
  (define (prune-subst subst env)
    (cond
      ((null? subst) '())
      ((not (assq (cdar subst) env)) (prune-subst (cdr subst) env))
      (else (cons (car subst) (prune-subst (cdr subst) env)))))
  (define (load file proc)
    (with-input-from-file file
       (lambda ()
         (let f ()
           (let ((x (read)))
             (unless (eof-object? x) 
               (proc x)
               (f)))))))
  (let-values (((code* subst env) (make-init-code)))
    (for-each
      (lambda (file)
        (display "expanding ")
        (display file)
        (newline)
        (load file
          (lambda (x)
            (let-values (((name code export-subst export-env)
                          (boot-library-expand x)))
               (set! code* (cons code code*))
               (set! subst (append export-subst subst))
               (set! env (append export-env env))))))
      files)
    (let-values (((export-subst export-env export-locs)
                  (make-system-data (prune-subst subst env) env)))
      (let ((code (build-system-library export-subst export-env export-locs)))
        (values
          (reverse (cons* (car code*) code (cdr code*)))
          export-locs)))))



(define bootstrap-collection
  (let ((ls '()))
    (case-lambda
      (() ls)
      ((x) 
       (unless (memq x ls) 
         (set! ls (cons x ls)))))))

(verify-map)


(let ((all-names (map car identifier->library-map))
      (all-labels (map (lambda (x) (gensym)) identifier->library-map))
      (all-bindings (map (lambda (x) 
                           (cond
                             ((macro-identifier x) => cadr)
                             (else `(core-prim . ,x))))
                         (map car identifier->library-map))))
  (let ((export-subst (map cons all-names all-labels))
        (export-env (map cons all-labels all-bindings)))
    (define (build-library legend-entry)
      (let ((key (car legend-entry))
            (name (cadr legend-entry))
            (visible? (caddr legend-entry))
            (required? (cadddr legend-entry)))
        (when required?
          (let ((id     (gensym))
                (name       name)
                (version     (if (eq? (car name) 'rnrs) '(6) '()))
                (import-libs '())
                (visit-libs  '())
                (invoke-libs '()))
            (let-values (((subst env)
                          (if (equal? name '(psyntax system $all))
                              (values export-subst export-env)
                              (values
                                (get-export-subset key export-subst)
                                '()))))
              (parameterize ((current-library-collection 
                              bootstrap-collection))
                (install-library
                   id name version import-libs visit-libs invoke-libs
                   subst env values values #f #f visible? #f)))))))
    (for-each build-library library-legend)))

(let ()
  (define-syntax define-prims
    (syntax-rules ()
      ((_ name* ...)
       (let ((g* (map (lambda (x) (gensym)) '(name* ...)))
             (v* (list name* ...)))
         (for-each set-symbol-value! g* v*)
         (let ((alist (map cons '(name* ...) g*)))
           (current-primitive-locations
             (lambda (x) 
               (cond
                 ((assq x alist) => cdr)
                 (else (error #f "undefined prim ~s" x))))))))))
  (define-prims
    syntax-dispatch apply cons append map list syntax-error
    generate-temporaries = + datum->syntax string->symbol
    string-append symbol->string syntax->datum gensym length 
    open-string-output-port identifier? free-identifier=? exists
    values call-with-values for-all null? cdr car pair? vector eq? bound-identifier=? reverse ellipsis-map assertion-violation))




(let-values (((core* locs)
               (parameterize ((current-library-collection bootstrap-collection))
                 (expand-all scheme-library-files))))
    (current-primitive-locations
      (lambda (x)
        (cond
          ((assq x locs) => cdr)
          (else #f))))
    (when (file-exists? "./psyntax.scm")
      (delete-file "./psyntax.scm"))
    (let ((p (open-output-file "./psyntax.scm")))
      (display ";;; Copyright (c) 2006, 2007 Abdulaziz Ghuloum and Kent Dybvig" p) (newline p)
      (display ";;; automatically generated from psyntax sources" p) (newline p)
      (display ";;; for copyright details, see psyntax/main.ss" p) (newline p) (newline p)
      (for-each
        (lambda (x)
          (compile-core-expr-to-port x p)
          (newline p))
        core*)
      (close-output-port p)
))

(display "Happy Happy Joy Joy\n")


;;; vim:syntax=scheme
